from pkginfo import SDist
from build import ProjectBuilder
import os


def get_pkg_metadata(filename):
    meta_pkg = SDist(filename)
    return meta_pkg


def build_package():
    # FIXME: Disable annoying warning
    # BetaConfiguration: Support for `[tool.setuptools]` in `pyproject.toml` is still *beta*.
    os.environ["PYTHONWARNINGS"] = "ignore"
    return ProjectBuilder(".").build("sdist", "dist")



import unittest
from whatsopt.mooptimization import MOOptimization, FLOAT, INT


class TestMOOptimization(unittest.TestCase):
    @unittest.skip("Need WhatsOpt connection, only used in development environment")
    def test_constructor(self):
        xtypes = [
            {"type": FLOAT, "limits": [0.0, 1.0]},
            {"type": INT, "limits": [0, 3]},
            {"type": INT, "limits": [0, 3]},
        ]

        _ = MOOptimization(xtypes, n_obj=3)



import os
import unittest
import subprocess
import json

dname = os.path.dirname


def file(name):
    return os.path.join(dname(__file__), "data", name)


COMMANDS = [
    "wop push -n {}".format(file("sellar.py")),
    "wop upload -n -a 1 {}".format(file("mda_init.py")),
    "wop upload -n {}".format(file("test_doe.csv")),
    "wop upload -n {}".format(file("test_doe.sqlite")),
    # skip as test data is obsolete
    # "wop upload -p -n {}".format(file("test_parallel_doe.sqlite_0")),
    "wop show -b -f {}".format(file("multipoint_beam/multipoint_beam_group.py")),
    "wop show -b --depth 3 -f {}".format(
        file("multipoint_beam/multipoint_beam_group.py")
    ),
    "wop status",
    "wop convert {}".format(file("test_doe.sqlite")),
]


def ordered(obj):
    if isinstance(obj, dict):
        return sorted((k, ordered(v)) for k, v in obj.items())
    if isinstance(obj, list):
        return sorted(ordered(x) for x in obj)
    else:
        return obj


class TestWopCommand(unittest.TestCase):
    def _test_wop_cmd(self, cmd):
        try:
            return subprocess.check_output(cmd.split(), encoding="utf-8")
        except subprocess.CalledProcessError as err:
            self.fail(
                "Command '{}' failed.  Return code: {}".format(cmd, err.returncode)
            )

    def test_commands(self):
        for cmd in COMMANDS:
            self._test_wop_cmd(cmd)
        if os.path.exists("test_doe.csv"):
            os.remove("test_doe.csv")

    def test_push_depth(self):
        self.maxDiff = None
        for d in range(3):
            print(f"DEPTH={d}")
            out = self._test_wop_cmd(
                "wop push -d {} -n {}".format(
                    d, file("multipoint_beam/multipoint_beam_group.py")
                )
            )
            with open(file("multipoint_beam_group_d{}.json".format(d))) as f:
                expected = ordered(json.loads(f.read()))
                actual = ordered(json.loads(out))
                self.assertTrue(expected, actual)


if __name__ == "__main__":
    unittest.main()



import unittest
import json

import openmdao.api as om
from openmdao.test_suite.test_examples.test_betz_limit import ActuatorDisc
from openmdao.test_suite.test_examples.beam_optimization.multipoint_beam_stress import (
    MultipointBeamGroup,
)
from whatsopt.push_command import PushCommand


def problem_init():
    prob = om.Problem()
    indeps = prob.model.add_subsystem("indeps", om.IndepVarComp(), promotes=["*"])
    indeps.add_output("a", 0.5)
    indeps.add_output("Area", 10.0, units="m**2")
    indeps.add_output("rho", 1.225, units="kg/m**3")
    indeps.add_output("Vu", 10.0, units="m/s")

    prob.model.add_subsystem("a_disk", ActuatorDisc(), promotes=["*"])

    # setup the optimization
    prob.driver = om.ScipyOptimizeDriver()
    prob.driver.options["optimizer"] = "SLSQP"

    prob.model.add_design_var("a", lower=0.0, upper=1.0)
    prob.model.add_design_var("Area", lower=0.0, upper=1.0)

    # negative one so we maximize the objective
    prob.model.add_objective("Cp", scaler=-1)

    prob.setup()
    prob.final_setup()
    return prob


def problem_init2():
    E = 1.0
    L = 1.0
    b = 0.1
    volume = 0.01
    max_bending = 100.0

    num_cp = 5
    num_elements = 25
    num_load_cases = 2

    model = MultipointBeamGroup(
        E=E,
        L=L,
        b=b,
        volume=volume,
        max_bending=max_bending,
        num_elements=num_elements,
        num_cp=num_cp,
        num_load_cases=num_load_cases,
    )

    prob = om.Problem(model=model)

    prob.driver = om.ScipyOptimizeDriver()
    prob.driver.options["optimizer"] = "SLSQP"
    prob.driver.options["tol"] = 1e-9
    prob.driver.options["disp"] = True

    prob.setup(mode="rev")
    prob.setup()
    prob.final_setup()
    return prob


class TestPushCommand2(unittest.TestCase):
    def test_get_mda_attributes(self):
        problem = problem_init2()
        push_cmd = PushCommand(problem, 0, False)
        mda_attrs = push_cmd.get_mda_attributes(problem.model, push_cmd.tree)
        print(json.dumps(mda_attrs, indent=2))


if __name__ == "__main__":
    unittest.main()



import numpy as np
from .optimization import Optimization

# from smt.applications.mixed_integer import FLOAT, INT, ORD, ENUM
# To avoid smt dependency replicate SMT mixed integer types
FLOAT = "float_type"
INT = "int_type"
ORD = "ord_type"
ENUM = "enum_type"

SEGMOOMOE = "SEGMOOMOE"


class OptimizationError(Exception):
    pass


class MOOptimization(Optimization):
    def __init__(
        self, xtypes=None, n_obj=2, cstr_specs=None, options=None, xlimits=None
    ):
        """Creation of a mixed-integer multi-objectives optimization context

        Parameters
        ----------
        xtypes: list of x inputs type specification (maybe optional if xlimits is specified, see below).
            A type specification consists in a dictionary of two items {type: ... , limits: ...} where type is either
            FLOAT, INT, ORD or ENUM and limits the corresponding bounds: for instance [-1.5, 3.0] for a FLOAT
            [1, 3] for an INT [1, 5, 7, 12] for an ORD and ["red", "green", "blue"] for an ENUM.
        n_obj: number of objectives (default=2)
        cstrs_specs: list of inequality constraints specifications (optional)
            A constraint specification is a dictionary of two items {type: ..., bound: ..., tol: ...}
            where type is either '<' or '>', bound is a float value specifying a limit and tol the
            acceptable tolerance before violation (ex if we want cstr_value < bound, we accept cstr_value < bound + tol)
        options: dictionary containing options to be passed to the mixed-integer multi-objectives optimizer
        xlimits: list of float limits in case we are dealing with float inputs
            Convenient way when x inputs are floats to specify only [[lower bound, upper bound], ...].

        Returns
        -------
        Optimizer handle allowing to trigger remote optimization operations

        """
        self._kind = SEGMOOMOE  # at the moment only one kind of MOO optimizer
        self._n_obj = n_obj
        self._cstr_specs = cstr_specs or []
        self._options = options

        if xtypes is None and xlimits is None:
            raise OptimizationError(
                "Configuration Error: you should specify either xtypes or xlimits"
            )

        if not (xtypes is None or xlimits is None):
            raise OptimizationError(
                "Configuration Error: you should specify either xtypes or xlimits exclusively, not both"
            )

        if xtypes is None and xlimits is not None:
            self._xtypes = []
            for xlim in xlimits:
                xlim = np.asarray(xlim).tolist()
                self._xtypes.append({"type": FLOAT, "limits": xlim})
            print(f"Convert {xlimits} to {self._xtypes}")
        else:
            self._xtypes = xtypes

        self._init_optimization()

    def _init_config(self):
        return {
            "kind": self._kind,
            "n_obj": self._n_obj,
            "xlimits": [],  # not used
            "xtypes": self._xtypes,
            "cstr_specs": self._cstr_specs,
            "options": self._options,
        }



import re
import os
import tempfile
from contextlib import contextmanager

from openmdao import __version__ as OPENMDAO_VERSION
from packaging.version import Version


def cut(mda_attrs, depth):
    if depth <= 0:
        return mda_attrs
    elif depth == 1:
        flatten(mda_attrs)
    else:
        for disc in mda_attrs["disciplines_attributes"]:
            sub_mdattrs = disc.get("sub_analysis_attributes")
            if sub_mdattrs:
                cut(sub_mdattrs, depth - 1)


def flatten(mda_attrs):
    # print("-------- flatten ", mda_attrs["name"])
    for disc in mda_attrs["disciplines_attributes"]:
        # print("************** ", disc["name"])
        sub_mdattrs = disc.get("sub_analysis_attributes")
        if sub_mdattrs:
            varattrs = disc.get("variables_attributes", [])
            subdriver = sub_mdattrs["disciplines_attributes"][0]
            # print("<<<< DRIVER ", subdriver["variables_attributes"])
            for vattr in subdriver["variables_attributes"]:
                v = vattr.copy()
                v["io_mode"] = "out" if vattr["io_mode"] == "in" else "in"
                varattrs.append(v)
            del disc["sub_analysis_attributes"]
            disc["variables_attributes"] = varattrs
            # print(">>>>", disc["name"])
            # print(disc["variables_attributes"])

            driver = mda_attrs["disciplines_attributes"][0]
            for vattr in subdriver["variables_attributes"]:
                already_present = [v["name"] for v in driver["variables_attributes"]]
                if vattr["name"] not in already_present:
                    v = vattr.copy()
                    # print("ADD DRIVER of ", mda_attrs["name"], v)
                    driver["variables_attributes"].append(v)


# push_command collect_var_infos
def format_shape(scalar, shape):
    shape = shape.replace("L", "")  # with py27 we can get (1L,)
    if scalar and shape == "(1,)":
        shape = "1"
    return shape


# push_command _get_sub_analysis_attributes & wop _get_discipline_attributes
def to_camelcase(name):
    return re.sub(r"(?:^|_)(\w)", lambda x: x.group(1).upper(), name)


# push_command get_mda_attributes
def simple_value(var):
    typ = var["type"]
    if var["shape"] == "1" or var["shape"] == "(1,)":
        ret = float(var["value"])
        if typ == "Integer":
            ret = int(ret)
    else:
        if typ == "Integer":
            var["value"] = var["value"].astype(int)
        else:
            var["value"] = var["value"].astype(float)
        ret = var["value"].tolist()
    return str(ret)


# push_command _get_varattr_from_connection
def extract_mda_var(fullname):
    name_elts = fullname.split(".")
    if len(name_elts) > 1:
        mda, var = name_elts[:-1], name_elts[-1]
    else:
        raise Exception(
            "Connection qualified name should contain"
            + " at least one dot, but got %s" % fullname
        )
    return mda, var


@contextmanager
def problem_pyfile(py_filename, component):
    dirname = os.path.dirname(py_filename)
    filename = os.path.basename(py_filename)
    module = os.path.splitext(filename)[0]
    content = """from openmdao.api import Problem, Group, IndepVarComp
from {} import {}

comp = {}()
comp_name = "{}"


class {}ComponentDiscovery(Group):
    pass


class {}Component(Group):
    pass


model_discovery = {}ComponentDiscovery()
model_discovery.add_subsystem(comp_name, comp, promotes=["*"])
pb_discovery = Problem(model_discovery)
pb_discovery.setup()


inputs = comp.list_inputs(out_stream=None)
indeps = IndepVarComp()
for name, meta in inputs:
    args = {{
        "val": meta.get("value", meta.get("val")),   # fix OpenMDAO < 3.10
        "shape": meta.get("shape"),
        "desc": meta.get("desc", ""),
        "units": meta.get("units", None)
    }}
    indeps.add_output(name, **args)

model = {}Component()
model.add_subsystem("indeps", indeps, promotes=["*"])
model.add_subsystem(comp_name, comp, promotes=["*"])
pb = Problem(model)
pb.setup()
pb.final_setup()
""".format(
        module,
        component,
        component,
        component,
        component,
        component,
        component,
        component,
    )
    handle, pbfile = tempfile.mkstemp(suffix=".py", dir=dirname)
    os.close(handle)
    with open(pbfile, "w") as pbf:
        pbf.write(content)

    try:
        yield pbfile
    finally:
        os.unlink(pbfile)


def find_indep_var_name(pb, absname):
    target = None
    for tgt, src in pb.model._conn_global_abs_in2out.items():
        if src == absname:
            target = tgt
            break

    name = None
    if Version(OPENMDAO_VERSION) < Version("3.39"):
        name = pb.model._var_abs2prom["input"].get(target)
    else:
        name = pb.model._resolver.abs2prom(target, "input")

    return name


def build_variable_name(varname1, varname2, limit=255):
    varname_sep = "=="
    ellipsis_sep = "..."

    name = varname1 + "==" + varname2
    ellipsed_name = name
    if len(name) <= limit or limit < len(varname_sep) + 2 * len(ellipsis_sep) + 4:
        ellipsed_name = name
    else:
        half_limit = (limit - len(varname_sep)) // 2
        ellipsed_name = (
            _ellipsize(varname1, half_limit)
            + varname_sep
            + _ellipsize(varname2, half_limit)
        )

    if len(ellipsed_name) > limit:
        ellipsed_name = ellipsis_sep + varname2[-(limit - len(ellipsis_sep)) :]

    return ellipsed_name


def _ellipsize(name, limit):
    ellipsis_sep = "..."

    half_limit = (limit - len(ellipsis_sep)) // 2
    if len(name) <= half_limit:
        return name
    else:
        return name[:half_limit] + ellipsis_sep + name[-half_limit:]



import os
import unittest
import tempfile

from whatsopt.utils import (
    WOP_CONF_FILENAME,
    is_analysis_user_file,
    is_based_on,
    is_user_file,
    find_analysis_base_files,
    get_analysis_id,
    get_whatsopt_url,
    load_state,
    save_state,
)


class TestUtils(unittest.TestCase):
    DATA_PATH = os.path.join(os.path.dirname(__file__), "data")
    MBEAM_PATH = os.path.join(DATA_PATH, "multipoint_beam")

    def setup(self):
        if os.path.exists(WOP_CONF_FILENAME):
            os.remove(WOP_CONF_FILENAME)

    def tearDown(self):
        if os.path.exists(WOP_CONF_FILENAME):
            os.remove(WOP_CONF_FILENAME)

    def test_save_load_state(self):
        state = {
            "wop_format_version": 2,
            "whatsopt_url": "https://example.com",
            "analysis_id": 666,
            "framework": "openmdao",
            "pull_mode": "package",
        }
        save_state(state)
        with open(WOP_CONF_FILENAME, "r") as file:
            data = file.read()
            with open(os.path.join(TestUtils.DATA_PATH, "wop_ref"), "r") as file:
                ref = file.read()
                self.assertEqual(ref, data)
        retrieved = load_state()
        self.assertEqual(state, retrieved)

    def test_load_state_old_format(self):
        state = {
            "wop_format_version": 1,
            "whatsopt_url": "https://erebe.onecert.fr/whatsopt",
            "analysis_id": 67,
            "framework": "gemseo",
            "pull_mode": "plain",
        }
        retrieved = load_state(
            filename=os.path.join(TestUtils.DATA_PATH, "wop_format_v1")
        )
        self.assertEqual(state, retrieved)

    def test_load_state_toml(self):
        state = {
            "wop_format_version": 2,
            "whatsopt_url": "https://erebe.onecert.fr/whatsopt",
            "analysis_id": 67,
            "framework": "gemseo",
            "pull_mode": "plain",
        }
        retrieved = load_state(
            filename=os.path.join(TestUtils.DATA_PATH, "wop_format_v2")
        )
        self.assertEqual(state, retrieved)

    def test_is_user_file(self):
        self.assertEqual(True, is_user_file("test.py"))
        self.assertEqual(False, is_user_file("test_base.py"))
        self.assertEqual(False, is_user_file("run_test.py"))

    def test_find_analysis_base_files(self):
        self.assertEqual(
            set(
                [
                    "i_comp_base.py",
                    "interp_base.py",
                    "multipoint_beam_group_base.py",
                    "volume_comp_base.py",
                    "obj_sum_base.py",
                    "local_stiffness_matrix_comp_base.py",
                ]
            ),
            set(find_analysis_base_files(TestUtils.MBEAM_PATH)),
        )

    def test_get_analysis_id(self):
        self.assertEqual("4", get_analysis_id(TestUtils.MBEAM_PATH))

    def test_get_whatsopt_url(self):
        self.assertEqual(
            "https://ether.onera.fr/whatsopt", get_whatsopt_url(TestUtils.MBEAM_PATH)
        )

    def test_based_on_openmdao(self):
        self.assertEqual(False, is_based_on("gemseo", TestUtils.MBEAM_PATH))
        self.assertEqual(True, is_based_on("openmdao", TestUtils.MBEAM_PATH))

    def test_based_on_framework_on_empty_dir(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            self.assertEqual(False, is_based_on("openmdao", tmpdir))
            self.assertEqual(False, is_based_on("gemseo", tmpdir))

    def test_is_analysis_user_file(self):
        self.assertEqual(True, is_analysis_user_file("sellar", "sellar.py"))
        self.assertEqual(False, is_analysis_user_file("sellar", "sellar_base.py"))


if __name__ == "__main__":
    unittest.main()



import os
import unittest
import numpy as np

from whatsopt.push_utils import (
    build_variable_name,
    simple_value,
    format_shape,
    to_camelcase,
    extract_mda_var,
    problem_pyfile,
)


class TestPushUtils(unittest.TestCase):
    DATA_PATH = os.path.join(os.path.dirname(__file__), "data")

    def test_format_shape(self):
        self.assertEqual("_file", format_shape(False, "L_file"))
        self.assertEqual("1", format_shape(True, "(1,)"))

    def test_to_camelcase(self):
        self.assertEqual("ToCamelCase", to_camelcase("to_camel_case"))
        self.assertEqual("Camel", to_camelcase("camel"))

    def test_simple_value(self):
        dict1 = {"type": "Integer", "shape": "1", "value": 1}
        dict2 = {"type": "Float", "shape": "(1,)", "value": 1.2}
        dict3 = {"type": "Float", "shape": None, "value": np.array(["1.2", "2.3"])}
        self.assertEqual("1", simple_value(dict1))
        self.assertEqual("1.2", simple_value(dict2))
        self.assertEqual("[1.2, 2.3]", simple_value(dict3))

    def test_extract_mda_var(self):
        val1, val2 = extract_mda_var("a.b.c.d")
        self.assertEqual(["a", "b", "c"], val1)
        self.assertEqual("d", val2)
        # raise Exception
        self.assertRaises(Exception, extract_mda_var, "a")

    def test_generate_problem_pyfile(self):
        with problem_pyfile(
            os.path.join(self.DATA_PATH, "disc1.py"), "Disc1"
        ) as pbfile:
            self.assertTrue(os.path.exists(pbfile))
            self.assertTrue(os.system("python {}".format(pbfile)) == 0)
        self.assertFalse(os.path.exists(pbfile))

    def test_build_varname(self):
        name = build_variable_name("12", "ab", 7)
        self.assertEqual("12==ab", name)
        name = build_variable_name("123456789", "abcdefghi", 11)
        self.assertEqual("...bcdefghi", name)
        name = build_variable_name("123456789", "abcdefghi", 12)
        self.assertEqual("1...9==a...i", name)
        name = build_variable_name("123456789", "abcdefghi", 13)
        self.assertEqual("1...9==a...i", name)
        name = build_variable_name("123456789", "abcdefghi", 14)
        self.assertEqual("1...9==a...i", name)
        name = build_variable_name("123456789", "abcdefghi", 15)
        self.assertEqual("1...9==a...i", name)
        name = build_variable_name("123456789", "abcdefghi", 16)
        self.assertEqual("12...89==ab...hi", name)
        name = build_variable_name("123456789", "abcdefghi", 17)
        self.assertEqual("12...89==ab...hi", name)
        name = build_variable_name("123456789", "abcdefghi", 18)
        self.assertEqual("12...89==ab...hi", name)
        name = build_variable_name("123456789", "abcdefghi", 19)
        self.assertEqual("12...89==ab...hi", name)
        name = build_variable_name("123456789", "abcdefghi", 20)
        self.assertEqual("123456789==abcdefghi", name)



import time
import numpy as np
from requests.exceptions import RequestException
from .whatsopt_client import WhatsOpt


SEGOMOE = "SEGOMOE"


class OptimizationError(Exception):
    pass


class Optimization:
    VALID_POINT = 0
    INVALID_POINT = 1
    RUNTIME_ERROR = 2
    SOLUTION_REACHED = 3
    RUNNING = 4
    PENDING = -1

    STATUSES = [
        "valid point",
        "invalid point",
        "runtime error",
        "solution reached",
        "running",
        "pending",
    ]

    DEFAULT_CSTR = {"type": "<", "bound": 0.0, "tol": 1e-4}
    TIMEOUT = 60

    def __init__(self, xlimits, cstr_specs=None, options=None):
        """Creation of a continuous mono objective optimization context

        Parameters
        ----------
        xlimits: list of float limits for x inputs eg [[lower bound, upper bound], ...]
        cstrs_specs: list of inequality constraints specifications (optional)
            A constraint specification is a dictionary of two items {type: ..., bound: ..., tol: ...}
            where type is either '<' or '>', bound is a float value specifying a limit and tol the
            acceptable tolerance before violation (ex if we want cstr_value < bound, we accept cstr_value < bound + tol).
        options: dictionary containing options to be passed to the remote optimizer

        Returns
        -------
        Optimizer handle allowing to trigger remote optimization operations

        """
        self._kind = SEGOMOE  # at the moment only one kind of Optimizer
        self._xlimits = xlimits or []
        self._cstr_specs = cstr_specs or []
        self._options = options
        self._init_optimization()

    def _init_optimization(self):
        try:
            for cstr in self._cstr_specs:
                for k in ["type", "bound", "tol"]:
                    cstr[k] = cstr.get(k, self.DEFAULT_CSTR[k])

            optim_config = self._init_config()

            self._wop = WhatsOpt().login()
            url = self._wop.endpoint("/api/v1/optimizations")
            resp = self._wop.session.post(
                url, headers=self._wop.headers, json={"optimization": optim_config}
            )
            if not resp.ok:
                message = "Unexpected error"
                if resp.json() and resp.json().get("message"):
                    message = resp.json()["message"]
                raise OptimizationError(
                    f"Error {resp.reason} ({resp.status_code}): {message}"
                )

            self._id = resp.json()["id"]
            self._x = np.array([])
            self._y = np.array([])
            self._x_suggested = None
            self._status = self.PENDING
            self._x_best = None
            self._y_best = None
        except RequestException:
            raise OptimizationError("Connection failed during initialization")

    def _init_config(self):
        return {
            "kind": self._kind,
            "n_obj": 1,
            "xlimits": self._xlimits,
            "cstr_specs": self._cstr_specs,
            "options": self._options,
        }

    def tell_doe(self, x, y):
        """Send the initial DOE to the remote optimizer

        This operation should be executed initially to set the initial data used
        to model objectives and constraints.

        Parameters
        ----------
        x : 2d array-like of inputs  [[x1, x2, ...], ...]
        y : 2d array-like of outputs [[obj1, obj2, ..., objn, cstr1, cstr2, ... cstrn], ...]
        """
        self._x_best = None
        self._status = self.PENDING
        self._x = np.atleast_2d(x)
        self._y = np.atleast_2d(y)
        if self._x.shape[0] != self._y.shape[0]:
            raise OptimizationError(
                f"Bad DOE error: DOE x and y should of the same size, got x size = {self._x.shape[0]} and  x size = {self._y.shape[0]}"
            )

    def run(self, f_grouped, n_iter=1, with_best=False):
        """Ask and tell the optimizer to get the optimum in n_iter iteration

        Parameters
        ----------
        f_grouped : function under optimization y = f(x)
            where y is of the form [obj1, obj2, ..., objn, cstr1, cstr2, ... cstrn]

        n_iter : int, iteration budget

        with_best : when set it compute the best point found at each iteration.

        Returns
        -------
        x_opt, y_opt: 2d array, 2d array
            an optimum in case of mono-objective or a list of optima in case
            of multi-objective optimization (Pareto front).
        """
        self._x_best = None
        for i in range(n_iter):
            with_best = with_best or (i == n_iter - 1)
            x_suggested, status, self._x_best, self._y_best = self.ask(with_best)
            print(
                f"{i} x suggested = {x_suggested} with status: {Optimization.STATUSES[status]}"
            )

            if with_best:
                print(f"x_best={self._x_best}")
                print(f"y_best={self._y_best}")

            # compute objective function at the suggested point
            new_y = f_grouped(x_suggested)
            print("new y = {}".format(new_y))

            self.tell(x_suggested, new_y)
            if self.is_solution_reached():
                print("Solution is reached")
                x_suggested, status, self._x_best, self._y_best = self.ask(
                    with_best=True
                )
                self._status = self.SOLUTION_REACHED
                break

        x_opt, y_opt = self.get_result()
        print(f"Found optimum y_opt = {y_opt} at x_opt = {x_opt}")
        return x_opt, y_opt

    def tell(self, x, y):
        """Gives (x, y) values such that y = f(x).

        Parameters
        ----------
        x : 2d array-like of inputs  [x1, x2, ...]
        y : 2d array-like of outputs [obj1, obj2, ..., objn, cstr1, cstr2, ... cstrn]
        """
        self._x = np.vstack((self._x, np.atleast_2d(x)))
        self._y = np.vstack((self._y, np.atleast_2d(y)))

    def ask(self, with_best=False):
        """Trigger optimizer iteration to get next promising location of the optimum

        Parameters
        ----------
        with_best: if set compute also the best (x_best, y_best) point so far among
        points already computed.
        In case of multi-objective the Pareto front is computed.

        Returns
        -------
        status, x_suggested, x_best, y_best
            where status is an status code (int) retuned by the optimizer
                  x_suggested the next promising optimum location
                  x_best, y_best the optima so far
        """
        self._status = self.RUNNING
        self._optimizer_iteration(with_best)
        retry = self.TIMEOUT
        url = self._wop.endpoint("/api/v1/optimizations/{}".format(self._id))
        resp = None
        while retry > 0 and self.is_running():
            resp = self._wop.session.get(url, headers=self._wop.headers)
            if resp.ok:
                result = resp.json()["outputs"]
                self._x_suggested = result["x_suggested"]
                self._status = result["status"]
                if with_best:
                    self._x_best = result["x_best"]
                    self._y_best = result["y_best"]
                else:
                    self._x_best = None
                    self._y_best = None
            elif resp.status_code == 400:
                message = "Unexpected error"
                if resp.json() and resp.json().get("message"):
                    message = resp.json()["message"]
                raise OptimizationError(
                    f"Error {resp.reason} ({resp.status_code}): {message}"
                )

            if self.is_running():
                if retry + 1 % 10 == 0:
                    print("Waiting for result...")
                time.sleep(1)
            retry = retry - 1

        if retry <= 0:
            self._x_suggested = None
            self._status = self.RUNTIME_ERROR
            self._x_best = None
            self._y_best = None
            raise OptimizationError(
                "Time out: please check status and history and may be ask again."
            )

        self._status = self.PENDING

        return self._x_suggested, self._status, self._x_best, self._y_best

    def get_result(self):
        """Retrieve optimum (or optima in case of moo) among the DOE with valid constraints

        Returns
        -------
            x, y where x and y are 2d ndarrays
        """
        if not self._x_best:
            _, _, self._x_best, self._y_best = self.ask(with_best=True)
        x_opt = np.array(self._x_best)
        y_opt = np.array(self._y_best)

        return x_opt, y_opt

    def get_history(self):
        """Retrieve optimization history (initial doe + (x suggestions, y))

        Returns
        -------
            x, y where x and y are 2d ndarrays
        """
        return self._x, self._y

    def get_status(self):
        """Retrieve status string"""
        return self.STATUSES[self._status]

    def is_solution_reached(self):
        return self._status == self.SOLUTION_REACHED

    def is_running(self):
        return self._status == self.RUNNING

    def _get_best_y(self, x):
        y = None
        x = np.array(x)
        for i in range(self._x.shape[0]):
            if np.equal(x, self._x[i]).all():
                y = self._y[i].tolist()
                break
        return y

    def _optimizer_iteration(self, with_best):
        """Run optimizer iteration"""
        if self._x.size == 0 or self._y.size == 0:
            raise OptimizationError(
                f"Empty DOE error: DOE should not be empty, got DOE x={self._x} and DOE y={self._y}"
            )
        if self._x.shape[0] != self._y.shape[0]:
            raise OptimizationError(
                f"Bad DOE error: DOE x and y should of the same size, got x size = {self._x.shape[0]} and  x size = {self._y.shape[0]}"
            )
        try:
            url = self._wop.endpoint("/api/v1/optimizations/{}".format(self._id))
            resp = self._wop.session.put(
                url,
                headers=self._wop.headers,
                json={
                    "optimization": {
                        "x": self._x.tolist(),
                        "y": self._y.tolist(),
                        "with_best": with_best,
                    }
                },
            )
            if not resp.ok:
                self._status = self.RUNTIME_ERROR
                raise OptimizationError(
                    f"Error {resp.reason} ({resp.status_code}): {resp.json().get('message', 'Unexpected Error')}"
                )
        except RequestException as e:
            raise OptimizationError(f"Connection failed: {e}")



import os
import unittest
from openmdao.api import CaseReader

from whatsopt.upload_utils import (
    load_from_csv,
    load_from_sqlite,
    load_from_hdf5,
    _format_upload_cases,
    _check_count,
)

GEMSEO_INSTALLED = True
try:
    import gemseo  # noqa: F401
except ImportError:
    GEMSEO_INSTALLED = False


class TestUploadUtils(unittest.TestCase):
    DATA_PATH = os.path.join(os.path.dirname(__file__), "data")

    def test_load_from_csv(self):
        filepath = os.path.join(TestUploadUtils.DATA_PATH, "test_doe.csv")
        name, cases, statuses = load_from_csv(filepath)
        self.assertEqual("test_doe", name)
        self.assertEqual("x", cases[0]["varname"])
        self.assertEqual(-1, cases[0]["coord_index"])
        self.assertEqual(7.5, cases[0]["values"][4])
        self.assertEqual(1, statuses[0])

    def test_load_from_sqlite(self):
        filepath = os.path.join(TestUploadUtils.DATA_PATH, "test_doe.sqlite")
        name, cases, statuses = load_from_sqlite(filepath)
        self.assertEqual("SMT_DOE_LHS", name)
        self.assertEqual(6.87783, round(cases[0]["values"][2], 5))
        self.assertEqual("z", cases[1]["varname"])
        self.assertEqual(1, cases[2]["coord_index"])
        for i in statuses:
            self.assertEqual(1, i)

    @unittest.skipUnless(GEMSEO_INSTALLED, "GEMSEO not installed")
    def test_load_from_hdf5(self):
        filepath = os.path.join(TestUploadUtils.DATA_PATH, "test_doe.hdf5")
        name, cases, statuses = load_from_hdf5(filepath)
        self.assertEqual("GEMSEO_DOE_ALGO", name)
        self.assertEqual(9.94343, round(cases[0]["values"][2], 5))
        self.assertEqual("z", cases[2]["varname"])
        self.assertEqual(1, cases[2]["coord_index"])
        for i in statuses:
            self.assertEqual(1, i)

    @unittest.skip(
        "Test data obsolete! Has to be regenerated running run_doe --parallel with openmdao MPI"
    )
    def test_load_from_parallel_sqlite(self):
        filepath = os.path.join(TestUploadUtils.DATA_PATH, "test_parallel_doe.sqlite_0")
        name, cases, statuses = load_from_sqlite(filepath)
        self.assertEqual("SMT_DOE_LHS", name)
        n = len(cases[0]["values"])
        self.assertEqual(
            50, n
        )  # supposedly run with 150 cases on 3 processors, hence 50
        self.assertEqual(n, len(statuses))

        _, cases2, statuses2 = load_from_sqlite(filepath, parallel=True)
        self.assertEqual(3 * n, len(cases2[0]["values"]))
        self.assertEqual(3 * n, len(statuses2))

        filepath1 = os.path.join(
            TestUploadUtils.DATA_PATH, "test_parallel_doe.sqlite_1"
        )
        _, cases2, statuses2 = load_from_sqlite(filepath1, parallel=True)
        self.assertEqual(2 * n, len(cases2[0]["values"]))
        self.assertEqual(2 * n, len(statuses2))

    def test_check_count(self):
        dict1 = {"A": [1, 2], "B": [3, 4], "C": [5, 6]}
        dict2 = {"A": [1, 2], "B": [3, 4], "C": [5, 6, 7]}
        self.assertEqual(2, _check_count(dict1))
        self.assertRaises(Exception, _check_count, dict2)

    def test_format_upload_cases(self):
        data, statuses = _format_upload_cases(
            CaseReader(os.path.join(TestUploadUtils.DATA_PATH, "test_doe.sqlite"))
        )
        print(data)
        self.assertEqual(6.87783, round(data[0]["values"][2], 5))
        self.assertEqual("z", data[1]["varname"])
        self.assertEqual(1, data[2]["coord_index"])
        for i in statuses:
            self.assertEqual(1, i)



import unittest
from whatsopt.whatsopt_client import WhatsOpt, EXTRANET_SERVER_URL


class WhatsOptClientTest(unittest.TestCase):
    def test_login_compat(self):
        # Check backward compatibility for FASTOAD
        wop = WhatsOpt(url=EXTRANET_SERVER_URL, login=False)
        self.assertEqual(EXTRANET_SERVER_URL, wop.url)



import os
import unittest

from whatsopt.convert_utils import (
    convert_sqlite_to_csv,
)


class TestUploadUtils(unittest.TestCase):
    DATA_PATH = os.path.join(os.path.dirname(__file__), "data")

    def test_convert_sqlite_to_csv(self):
        filepath = os.path.join(TestUploadUtils.DATA_PATH, "test_doe.sqlite")
        convert_sqlite_to_csv(filepath, "test_doe")
        self.assertTrue(os.path.exists("test_doe.csv"))
        os.remove("test_doe.csv")



import re
from whatsopt.push_utils import (
    build_variable_name,
    cut,
    find_indep_var_name,
    simple_value,
    extract_mda_var,
    format_shape,
)
from whatsopt.logging import debug
from openmdao.api import IndepVarComp

from openmdao.visualization.n2_viewer.n2_viewer import _get_viewer_data

from openmdao import __version__ as OPENMDAO_VERSION
from packaging.version import Version

# Special name for internal WhatsOpt discipline. cf. WhatsOpt Discipline model
DRIVER_NAME = "__DRIVER__"
# OpenMDAO 3.2+ component name (handles indep vars automatically)
AUTO_IVC = "_auto_ivc"


class PushCommand:
    """
    This push command allows to push any OpenMDAO problem
    (as opposed to regular push command which works with "all vars promoted/no connect" assumption)
    """

    def __init__(self, problem, depth, scalar):
        data = _get_viewer_data(problem)
        self.problem = problem
        self.depth = depth
        self.scalar = scalar
        self.tree = data["tree"]
        self.connections = data["connections_list"]
        self.vars = {"in": {}, "out": {}}
        self.vardescs = {}
        self.discmap = {}
        self.mdas = {}

    def get_mda_attributes(self, group, tree, use_depth=False):
        self._collect_disc_infos(self.problem.model, self.tree)
        self._collect_var_infos(self.problem)

        mda_attrs = self._get_mda_hierarchy(group, tree)

        mda_attrs["name"] = ""  # root -> ""
        self._populate_varattrs_from_connections(mda_attrs)
        self._populate_varattrs_from_outputs(mda_attrs)
        self._populate_initial_values(mda_attrs)
        for _, mda in self.mdas.items():
            self._populate_initial_values(mda)

        mda_attrs["name"] = group.__class__.__name__

        if use_depth:
            cut(mda_attrs, self.depth)

        return mda_attrs

    def _get_mda_hierarchy(self, group, tree, group_prefix=""):
        name = tree["name"]

        mda_attrs = {
            "name": name,
            "disciplines_attributes": [
                {"name": DRIVER_NAME, "variables_attributes": []}
            ],
        }
        self.mdas[group_prefix] = mda_attrs
        for child in tree["children"]:
            for s in group._subsystems_myproc:
                if s.name == child["name"]:
                    if (
                        child["type"] == "subsystem"
                        and child["subsystem_type"] == "group"
                    ):
                        prefix = child["name"]
                        if group_prefix:
                            prefix = group_prefix + "." + child["name"]
                        sub_analysis_attrs = self._get_sub_analysis_attributes(
                            s, child, prefix
                        )
                        mda_attrs["disciplines_attributes"].append(sub_analysis_attrs)
                    else:
                        if not isinstance(s, IndepVarComp):
                            discattrs = {
                                "name": child["name"],
                                "variables_attributes": [],
                            }
                            mda_attrs["disciplines_attributes"].append(discattrs)

        return mda_attrs

    def _populate_varattrs_from_connections(self, mda_attrs):
        for conn in self.connections:
            mda_src, varname_src = extract_mda_var(conn["src"])
            mda_tgt, varname_tgt = extract_mda_var(conn["tgt"])
            hat = []
            while mda_src and mda_tgt and mda_src[0] == mda_tgt[0]:
                hat.append(mda_src[0])
                mda_src = mda_src[1:]
                mda_tgt = mda_tgt[1:]
            conn_name = self._get_conn_name(conn)
            for discattrs in self.mdas[".".join(hat)]["disciplines_attributes"]:
                # mda hat: src
                if (
                    mda_src[0] != AUTO_IVC
                    and self.discmap[mda_src[0]] == discattrs["name"]
                ):
                    self._set_varattr(
                        discattrs, conn["src"], varname_src, conn_name, "out"
                    )
                    # src -> hat
                    self._set_varattr_in_depth(
                        discattrs.get("sub_analysis_attributes"),
                        mda_src[1:],
                        conn["src"],
                        conn_name,
                        upward=True,
                    )

                # mda hat: tgt
                if self.discmap[mda_tgt[0]] == discattrs["name"]:
                    self._set_varattr(
                        discattrs, conn["tgt"], varname_tgt, conn_name, "in"
                    )
                    if mda_src[0] == AUTO_IVC:
                        driver_attrs = mda_attrs["disciplines_attributes"][0]
                        self._set_varattr(
                            driver_attrs, conn["src"], varname_tgt, conn_name, "out"
                        )

                    # hat -> tgt
                    self._set_varattr_in_depth(
                        discattrs.get("sub_analysis_attributes"),
                        mda_tgt[1:],
                        conn["tgt"],
                        conn_name,
                        upward=False,
                    )

    def _populate_varattrs_from_outputs(self, mda_attrs, group_prefix=""):
        mda_prefix = group_prefix
        if mda_prefix:
            mda_prefix += "."
        if mda_attrs["name"]:
            mda_prefix += mda_attrs["name"] + "."
        for discattrs in mda_attrs["disciplines_attributes"]:
            sub_mda_attrs = discattrs.get("sub_analysis_attributes")
            if sub_mda_attrs is None:
                for absname, varattrs in self.vars["out"].items():
                    mda, _ = extract_mda_var(absname)
                    scope = discattrs["name"]
                    if mda_prefix:
                        scope = mda_prefix + scope
                    mda_name = ".".join(mda)
                    if mda_name == scope or (
                        mda_name.startswith(scope) and mda_name[len(scope)] == "."
                    ):
                        vattr = {
                            "name": varattrs["name"],
                            "desc": self.vardescs.get(absname, ""),
                            "type": varattrs["type"],
                            "shape": varattrs["shape"],
                            "units": varattrs["units"],
                        }
                        self._set_varattrs_from_outputs(
                            vattr, "out", discattrs["variables_attributes"]
                        )
                        dvattr = vattr.copy()
                        dvattr["io_mode"] = "in"
                        driver_attrs = mda_attrs["disciplines_attributes"][0]
                        self._set_varattrs_from_outputs(
                            dvattr, "in", driver_attrs["variables_attributes"]
                        )
            else:
                self._populate_varattrs_from_outputs(sub_mda_attrs, mda_prefix[:-1])

    def _populate_initial_values(self, mda_attrs):
        driver_attrs = mda_attrs["disciplines_attributes"][0]
        for vattr in driver_attrs["variables_attributes"]:
            if vattr["io_mode"] == "out":
                # set init value for design variables and parameters (outputs of driver)
                conn_name = vattr["name"].split("==")[0]
                if self.vars["out"].get(conn_name):
                    v = self.vars["out"][conn_name]
                    vattr["parameter_attributes"] = {"init": simple_value(v)}
                else:  # indep comp promoted
                    for _, v in self.vars["out"].items():
                        if conn_name == v["name"]:
                            vattr["parameter_attributes"] = {"init": simple_value(v)}
                            break

    def _set_varattr_in_depth(
        self, mda_attrs, mda_endpoint, endpoint, conn_name, upward
    ):
        if mda_attrs:
            io_mode = "out" if upward else "in"
            var = self.vars[io_mode][endpoint]
            varattrs = {
                "name": conn_name,
                "io_mode": io_mode,
                "desc": self.vardescs.get(endpoint, ""),
                "type": var["type"],
                "shape": var["shape"],
                "units": var["units"],
            }
            disc_endpoint = mda_endpoint[0]
            for discattrs in mda_attrs["disciplines_attributes"]:
                if self.discmap[disc_endpoint] == discattrs["name"]:
                    if discattrs.get("variables_attributes") is not None:
                        already_present = [
                            varattr["name"]
                            for varattr in discattrs["variables_attributes"]
                        ]
                        if conn_name not in already_present:
                            discattrs["variables_attributes"].append(varattrs)
                    submda_attrs = discattrs.get("sub_analysis_attributes")
                    break
            driver_varattrs = mda_attrs["disciplines_attributes"][0][
                "variables_attributes"
            ]
            already_present = [varattr["name"] for varattr in driver_varattrs]
            if conn_name not in already_present:
                varattr_driver = varattrs.copy()
                varattr_driver["io_mode"] = "in" if upward else "out"
                driver_varattrs.append(varattr_driver)
            self._set_varattr_in_depth(
                submda_attrs, mda_endpoint[1:], endpoint, conn_name, upward
            )

    def _set_varattr(self, discattrs, endpoint, varname, conn_name, io_mode):
        if discattrs.get("variables_attributes") is None:
            return
        already_present = [
            varattr["name"] for varattr in discattrs["variables_attributes"]
        ]
        if conn_name not in already_present:
            var = self.vars[io_mode][endpoint]
            discattrs["variables_attributes"].append(
                {
                    "name": conn_name,
                    "io_mode": io_mode,
                    "desc": self.vardescs.get(endpoint, ""),
                    "type": var["type"],
                    "shape": var["shape"],
                    "units": var["units"],
                }
            )

    def _set_varattrs_from_outputs(self, varattr, io_mode, varattrs):
        already_present = [varattr["name"] for varattr in varattrs]
        if varattr["name"] not in already_present:
            vattr = varattr.copy()
            vattr["io_mode"] = io_mode
            varattrs.append(vattr)

    def _get_conn_name(self, conn):
        if (
            self.vars["out"][conn["src"]]["name"]
            == self.vars["in"][conn["tgt"]]["name"]
        ):
            return self.vars["out"][conn["src"]]["name"]
        elif conn["src"].startswith(
            AUTO_IVC
        ):  # special case take varname of target when auto_ivc
            _, varname = extract_mda_var(conn["tgt"])
            return varname
        else:
            return build_variable_name(conn["src"], conn["tgt"], limit=255)

    def _get_sub_analysis_attributes(self, group, child, prefix):
        submda_attrs = self._get_mda_hierarchy(group, child, prefix)
        submda_attrs["name"] = child["name"]
        submda_attrs["disciplines_attributes"] = submda_attrs["disciplines_attributes"]
        self.mdas[prefix] = submda_attrs
        superdisc_attrs = {
            "name": child["name"],
            # "variables_attributes": [],
            "sub_analysis_attributes": submda_attrs,
        }
        return superdisc_attrs

    # see _get_tree_dict at
    # https://github.com/OpenMDAO/OpenMDAO/blob/master/openmdao/visualization/n2_viewer/n2_viewer.py
    def _collect_disc_infos(self, system, tree):
        if "children" not in tree:
            return

        for child in tree["children"]:
            for s in system._subsystems_myproc:
                if s.name == child["name"]:
                    if (
                        child["type"] == "subsystem"
                        and child["subsystem_type"] == "group"
                    ):
                        self.discmap[child["name"]] = child["name"]
                        self._collect_disc_infos(s, child)
                    else:
                        # do not represent IndepVarComp
                        if isinstance(s, IndepVarComp):
                            self.discmap[child["name"]] = DRIVER_NAME
                        else:
                            self.discmap[child["name"]] = child["name"]

    # see _get_tree_dict at
    # https://github.com/OpenMDAO/OpenMDAO/blob/master/openmdao/visualization/n2_viewer/n2_viewer.py
    def _collect_var_infos(self, problem):
        system = problem.model
        for io in ("input", "output"):
            for abs_name in system._var_abs2meta[io]:
                if io == "input":
                    io_mode = "in"
                elif io == "output":
                    io_mode = "out"
                else:
                    raise Exception("Unhandled variable type " + io)
                meta = system._var_abs2meta[io][abs_name]

                vtype = "Float"
                val = meta.get("value", meta.get("val"))  # fix OpenMDAO < 3.10
                if re.match("int", type(val).__name__):
                    vtype = "Integer"
                shape = str(meta["shape"])
                shape = format_shape(self.scalar, shape)

                name = None
                if Version(OPENMDAO_VERSION) < Version("3.39"):
                    name = system._var_abs2prom[io][abs_name]
                else:
                    name = system._resolver.abs2prom(abs_name, io)

                # name = abs_name
                if abs_name.startswith(AUTO_IVC):
                    name = find_indep_var_name(problem, abs_name)
                self.vars[io_mode][abs_name] = {
                    "name": name,
                    "type": vtype,
                    "shape": shape,
                    "units": meta["units"],
                    "value": val,
                }

                # retrieve initial conditions
                var = self.vars[io_mode][abs_name]
                if abs_name in system._outputs._views:
                    var["value"] = system._outputs[abs_name]
                elif abs_name in system._inputs._views:
                    var["value"] = system._inputs[abs_name]
                elif abs_name in system._discrete_outputs:
                    var["value"] = system._discrete_outputs[abs_name]
                elif abs_name in system._discrete_inputs:
                    var["value"] = system._discrete_inputs[abs_name]

                desc = self.vardescs.setdefault(name, "")
                if desc == "":
                    self.vardescs[name] = meta["desc"]
                elif desc != meta["desc"] and meta["desc"] != "":
                    debug(
                        'Find another description for {}: "{}", keep "{}"'.format(
                            name, meta["desc"], self.vardescs[name]
                        )
                    )



import csv
import openmdao.api as om
from whatsopt.logging import log

from whatsopt.upload_utils import load_sqlite_file


def convert_sqlite_to_csv(sqlite_filename, basename):
    cr = om.CaseReader(sqlite_filename)
    driver_cases = cr.list_cases("driver", out_stream=None)
    design_vars = sorted(cr.get_case(driver_cases[0]).get_design_vars())
    driver_name, cases, statuses = load_sqlite_file(sqlite_filename)
    all_vars = [v["varname"] for v in cases]
    out_vars = sorted(list(set(all_vars) - set(design_vars)))

    fieldnames = []
    values = {}
    nb_cases = len(statuses)
    for name in design_vars + out_vars:
        for case in cases:
            if name == case["varname"]:
                if case["coord_index"] == -1:
                    fieldnames.append(name)
                    values[name] = case["values"]
                else:
                    name_i = f"{name}[{case['coord_index']}]"
                    fieldnames.append(name_i)
                    values[name_i] = case["values"]

    outfile = f"{basename}.csv"
    with open(outfile, "w") as f:
        writer = csv.writer(f, delimiter=";", lineterminator="\n")
        writer.writerow(["success"] + fieldnames)

        for i in range(nb_cases):
            data = [statuses[i]]
            for var in fieldnames:
                data.append(values[var][i])
            writer.writerow(data)
    log(f"Convert {nb_cases} cases ({driver_name}) to {outfile}")



from http import HTTPStatus
import os
import sys
import json
import getpass
import requests
import copy
import re
import zipfile
import tempfile
import numpy as np
import time
from datetime import datetime
import tomli
import tomli_w
from tabulate import tabulate
from urllib.parse import urlparse
from packaging.version import Version

# push
import openmdao.utils.hooks as hooks
from openmdao.utils.file_utils import _load_and_exec

from openmdao.utils.webview import webview
from openmdao.api import IndepVarComp

from openmdao import __version__ as OPENMDAO_VERSION

from whatsopt.convert_utils import convert_sqlite_to_csv

from whatsopt.logging import log, info, warn, error, debug
from whatsopt.publish_utils import build_package, get_pkg_metadata
from whatsopt.utils import (
    FRAMEWORK_GEMSEO,
    FRAMEWORK_OPENMDAO,
    MODE_PACKAGE,
    MODE_PLAIN,
    extract_remote_name,
    is_analysis_user_file,
    is_based_on,
    is_framework_switch,
    is_package_mode,
    is_run_script_file,
    is_test_file,
    is_user_file,
    get_analysis_id,
    get_whatsopt_url,
    move_files,
    save_state,
)
from whatsopt.upload_utils import (
    load_from_csv,
    load_from_sqlite,
    load_from_hdf5,
    print_cases,
)
from whatsopt.push_utils import (
    find_indep_var_name,
    problem_pyfile,
    to_camelcase,
)
from whatsopt.push_command import PushCommand
from whatsopt.show_utils import generate_xdsm_html

from whatsopt import __version__

WHATSOPT_DIRNAME = os.path.join(os.path.expanduser("~"), ".whatsopt")
API_KEY_FILENAME = os.path.join(WHATSOPT_DIRNAME, "api_key")
URL_FILENAME = os.path.join(WHATSOPT_DIRNAME, "url")
REMOTES_FILENAME = os.path.join(WHATSOPT_DIRNAME, "remotes")

EXTRANET_SERVER_URL = "https://ether.onera.fr/whatsopt"


class WhatsOptImportMdaError(Exception):
    pass


class AnalysisPushedException(Exception):
    def __init__(self, xdsm):
        self.xdsm = xdsm


class WhatsOpt:
    def __init__(self, url=None, api_key=None, login=None):
        self._remotes = self._read_remotes()

        remote = self._remotes.get(url)
        debug(f"remote={remote}")
        if remote:
            self._url = remote["url"]
            self._api_key = api_key if api_key else remote["api_key"]
        else:
            debug(f"url={url} logged?={self.is_logged()}")
            if url:
                self._url = url.strip("/")
                current_url = self._read_url()
                if current_url and self._url != current_url:
                    self.logout(echo=False)
            elif self.is_logged():
                self._url = self._read_url()
            else:
                self._url = EXTRANET_SERVER_URL

            if api_key:
                self._api_key = api_key
            else:
                self._api_key = self._read_api_key()

        # config session object
        self.session = requests.Session()
        urlinfos = urlparse(self._url)
        self.session.trust_env = re.match(r"\w+.onera\.fr", urlinfos.netloc)
        self.headers = {}

    @property
    def url(self):
        return self._url

    @property
    def api_key(self):
        return self._api_key

    def endpoint(self, path):
        return self._url + path

    @staticmethod
    def _read_remotes():
        remotes = {}
        if os.path.exists(REMOTES_FILENAME):
            with open(REMOTES_FILENAME, "rb") as f:
                remotes = tomli.load(f)
        return remotes

    @staticmethod
    def _write_remotes(remotes):
        with open(REMOTES_FILENAME, "wb") as f:
            tomli_w.dump(remotes, f)

    def _ask_api_key(self):
        log("You have to set your API key.")
        log("You can get it in your profile page on WhatsOpt (%s)." % self.url)
        info(
            "Please, copy/paste your API key below then hit return (characters are hidden)."
        )
        api_key = getpass.getpass(prompt="Your API key: ")
        return api_key

    @staticmethod
    def is_logged():
        # Defensive programming: test on api_key should be enough but...
        return os.path.exists(URL_FILENAME) and os.path.exists(API_KEY_FILENAME)

    @staticmethod
    def _read_url():
        if os.path.exists(URL_FILENAME):
            with open(URL_FILENAME, "r") as f:
                return f.read()
        else:
            return None

    @staticmethod
    def _read_api_key():
        if os.path.exists(API_KEY_FILENAME):
            with open(API_KEY_FILENAME, "r") as f:
                return f.read()
        else:
            return None

    def _write_login_infos(self):
        if not os.path.exists(WHATSOPT_DIRNAME):
            os.makedirs(WHATSOPT_DIRNAME)
        with open(API_KEY_FILENAME, "w") as f:
            f.write(self.api_key)
        with open(URL_FILENAME, "w") as f:
            f.write(self.url)

    def login(self, echo=False, retry=True):
        debug(f"login(api_key={self.api_key}, echo={echo})")

        # Check login with remote name
        if not (
            self.url.startswith("http://") or self.url.startswith("https://")
        ) and not self._remotes.get(self.url):
            error(f"Unknown remote server '{self.url}'")
            WhatsOpt.list_remotes()
            sys.exit(-1)

        if self._api_key:
            # api key is provided
            pass
        elif self.is_logged():
            # check if logged in
            self._api_key = self._read_api_key()
        else:
            # check if known remote
            for _, v in self._remotes.items():
                if self.url == v["url"]:
                    self._api_key = v["api_key"]
                    break
            # ask for API key otherwise
            debug("Ask for API key")
            self._api_key = self._ask_api_key() if not self._api_key else self._api_key

        debug(f"url={self.url}, api_key={self.api_key}")
        ok = self._test_connection()
        debug(f"ok={ok}")
        if not ok and retry:
            # try to log again
            # log out silently as one may be logged on another server
            self.logout(echo=False)
            self._remotes = {
                k: v for k, v in self._remotes.items() if v["url"] != self.url
            }
            # save url again as is has been wipe out by logout
            with open(URL_FILENAME, "w") as f:
                f.write(self._url)
            return self.login(retry=False)

        # make remote info
        if ok:
            self._write_login_infos()
            remote_name = extract_remote_name(self.url)
            if remote_name:
                self._remotes[remote_name] = {"url": self.url, "api_key": self.api_key}

        debug(self._remotes)
        self._write_remotes(self._remotes)

        if not ok:
            error("Login to WhatsOpt ({}) failed.".format(self.url))
            log("")
            warn("You are not connected.")
            info("  use `wop login <url>` to connect to a remote WhatsOpt server")
            info(
                "  use `wop login <name>` to connect to a known remote WhatsOpt server"
            )
            self.list_remotes()

            sys.exit(-1)

        if echo:
            info(f"Successfully logged in to remote WhatsOpt {self.url}")
            log("")
        return self

    @staticmethod
    def logout(list=None, all=None, remote=None, echo=True):
        if list:
            WhatsOpt.list_remotes()
        elif all:
            WhatsOpt._write_remotes({})
            if echo:
                info("Sucessfully logged out from all WhatsOpt remotes")
        elif remote:
            remotes = WhatsOpt._read_remotes()
            if remotes.get(remote):
                del remotes[remote]
                WhatsOpt._write_remotes(remotes)
                if echo:
                    info(f"Sucessfully logged out from remote WhatsOpt {remote}")
        else:
            if os.path.exists(API_KEY_FILENAME):
                os.remove(API_KEY_FILENAME)
            url = WhatsOpt._read_url()
            if url:
                os.remove(URL_FILENAME)
            if echo:
                if url:
                    info(f"Sucessfully logged out from remote WhatsOpt {url}")
                else:
                    info("Not connected. You may want to connect to:")
                    WhatsOpt.list_remotes()

        if echo:
            log("")

    @staticmethod
    def list_remotes():
        remotes = WhatsOpt._read_remotes()
        headers = ["name", "url"]
        data = []
        for name, rem in remotes.items():
            data.append([name, rem["url"]])
        info("Known remote servers")
        log(tabulate(data, headers))

    def list_analyses(self, all=False, project_query=None):
        param = ""
        if all:
            param = "?all=true"
        elif project_query:
            param = "?design_project_query={}".format(project_query)
        url = self.endpoint("/api/v1/analyses" + param)
        resp = self.session.get(url, headers=self.headers)
        if resp.ok:
            mdas = resp.json()
            headers = ["id", "name", "created at"]
            data = []
            for mda in mdas:
                date = mda.get("created_at", None)
                data.append([mda["id"], mda["name"], date])
            log(tabulate(data, headers))
            log("")
        else:
            WhatsOpt.check_http_error(resp)

    def is_connected(self):
        return self._test_connection()

    def get_status(self):
        connected = self.is_connected()
        whatsopt_url = get_whatsopt_url() or self.url
        if connected:
            info("You are logged in {}".format(self.url))
        else:
            info("You are not connected.")
        mda_id = None
        try:
            mda_id = get_analysis_id()
        except ValueError as err:
            warn(str(err))
        if mda_id:
            if connected and whatsopt_url == self.url:
                info("Found local analysis code (id=#{})".format(mda_id))
                # connected to the right server from which the analysis was pulled
                url = self.endpoint("/api/v1/analyses/{}".format(mda_id))
                resp = self.session.get(url, headers=self.headers)

                if resp.ok:
                    mda = resp.json()
                    if is_based_on(FRAMEWORK_GEMSEO):
                        mda["framework"] = "GEMSEO"
                    elif is_based_on(FRAMEWORK_OPENMDAO):
                        mda["framework"] = "OpenMDAO"
                    else:  # should not happen
                        raise ValueError(
                            "No framework detected. Check your *_base.py files."
                        )
                    headers = ["id", "name", "created_at", "owner_email", "framework"]
                    data = [[mda[k] for k in headers]]
                    log(tabulate(data, headers))
                else:
                    error("Analysis not found on the server anymore (probably deleted)")
                    log(
                        "  (use 'wop push <analysis.py>' to push from an OpenMDAO code to the server)"
                    )
            else:
                info(
                    "Found local analysis code (id=#{}) pulled from {}".format(
                        mda_id, whatsopt_url
                    )
                )
                if connected:
                    # connected to another server with a pulled analysis
                    warn("You are connected to a different server")
                    log(
                        "  (use 'wop push <analysis.py>' to push from the local "
                        f"analysis code to the server {self.url})"
                    )
                    log(
                        f"  (use 'wop logout' and 'wop login {whatsopt_url}' "
                        "to log in to the original server)"
                    )
                else:
                    log(f"  (use 'wop login {whatsopt_url}' command to log in)")
        else:
            info("No local analysis found")
            if connected:
                log(
                    "  (use 'wop list' and 'wop pull <id>' to retrieve an existing analysis)\n"
                    "  (use 'wop push <analysis.py>' to push from the local OpenMDAO code to the server)"
                )
        log("")

    def push_component_cmd(self, py_filename, component, options):
        with problem_pyfile(py_filename, component) as pyf:
            if options["--dry-run"]:
                with open(pyf, "r") as pbf:
                    print(pbf.read())
                    sys.exit()
            self.push_mda_cmd(pyf, options)

    def push_mda_cmd(self, py_filename, options):
        def push_mda(prob):
            name = options["--name"]
            pbname = prob.model.__class__.__name__
            if name and pbname != name:
                info("Analysis %s skipped" % pbname)
                # do not exit seeking for another problem (ie analysis)
            else:
                options["--pyfilename"] = py_filename
                xdsm = self.push_mda(prob, options)
                if options.get("--xdsm"):  # show command
                    # required to interrupt pb execution
                    raise AnalysisPushedException(xdsm=xdsm)
                else:
                    sys.exit()

        hooks.use_hooks = True
        hooks._register_hook("final_setup", "Problem", post=push_mda)
        _load_and_exec(py_filename, [])
        return push_mda

    def push_mda(self, problem, options):
        scalar = options.get("--scalar")
        depth = options.get("--depth")
        push_cmd = PushCommand(problem, depth, scalar)

        mda_attrs = push_cmd.get_mda_attributes(
            problem.model, push_cmd.tree, use_depth=True
        )

        if mda_attrs["name"] == "Group" and options.get("--pyfilename"):
            mda_attrs["name"] = os.path.splitext(
                to_camelcase(os.path.basename(options.get("--pyfilename")))
            )[0]

        if options["--dry-run"]:
            log(json.dumps(mda_attrs, indent=2))
        else:
            suffix = ""
            if options.get("--xdsm"):
                suffix = ".xdsm"
            url = self.endpoint("/api/v1/analyses{}".format(suffix))
            resp = self.session.post(
                url, headers=self.headers, json={"analysis": mda_attrs}
            )
            WhatsOpt.check_http_error(resp)
            log("Analysis %s pushed" % mda_attrs["name"])
            return resp.json()

    def push_json(self, filename):
        with open(filename, "rb") as f:
            attrs = json.load(f)
        if "analyses_attributes" in attrs:  # project detection
            url = self.endpoint("/api/v1/design_projects")
            key = "Project"
        else:
            url = self.endpoint("/api/v1/analyses")
            key = "Analysis"
        params = {}
        params[key.lower()] = attrs
        resp = self.session.post(url, headers=self.headers, json=params)
        WhatsOpt.check_http_error(resp)
        log("{} {} pushed".format(key, attrs["name"]))

    def pull_mda(self, mda_id, options={}, msg=None, info_keep_run_ops=True):
        if not msg:
            msg = "Analysis %s pulled" % mda_id

        framework = FRAMEWORK_OPENMDAO
        if options.get("--gemseo"):
            framework = FRAMEWORK_GEMSEO

        param = ""
        if options.get("--run-ops"):
            param += "&with_runops=true"
        if options.get("--server"):
            if framework == FRAMEWORK_OPENMDAO:
                param += "&with_server=true"
            else:
                warn(
                    "Can not generate server with GEMSEO framework. --server is ignored"
                )
        if options.get("--egmdo"):
            if framework == FRAMEWORK_OPENMDAO:
                param += "&with_egmdo=true"
            else:
                warn("Can not generate EGMDO with GEMSEO framework. --egmdo is ignored")
        if options.get("--test-units"):
            if framework == FRAMEWORK_OPENMDAO:
                param += "&with_unittests=true"
            else:
                warn(
                    "Can not generate tests with GEMSEO framework. --test-units is ignored"
                )
        if param:
            param = "?" + param[1:]

        format_query = framework
        if options.get("--package"):
            format_query += "_pkg"

        url = self.endpoint(
            ("/api/v1/analyses/{}/exports/new.{}{}".format(mda_id, format_query, param))
        )
        resp = self.session.get(url, headers=self.headers, stream=True)
        WhatsOpt.check_http_error(resp)
        name = None
        with tempfile.NamedTemporaryFile(suffix=".zip", mode="wb", delete=False) as fd:
            for chunk in resp.iter_content(chunk_size=128):
                fd.write(chunk)
            name = fd.name
        zipf = zipfile.ZipFile(name, "r")
        tempdir = tempfile.mkdtemp(suffix="wop", dir=tempfile.tempdir)
        zipf.extractall(tempdir)
        filenames = zipf.namelist()
        zipf.close()
        file_to_move = {}
        if options.get("--dry-run"):
            # cmd = "Pull"
            # if options.get("--update"):
            #     cmd = "Update"
            info(
                "***************************************************\n"
                "* DRY RUN mode (actions are listed but not done) *\n"
                "***************************************************"
            )

        # Special case: When framework switch we have to update
        # user analysis main file which name depends on package name
        framework_switch = is_framework_switch(framework)
        if framework_switch:
            url = self.endpoint(f"/api/v1/analyses/{mda_id}/openmdao_impl")
            resp = self.session.get(url, headers=self.headers, stream=True)
            WhatsOpt.check_http_error(resp)
            mda_name = resp.json()["packaging"]["package_name"]
        else:
            mda_name = ""

        for f in filenames:
            file_to = f
            file_to_move[file_to] = True
            if os.path.exists(file_to):
                if options.get("--force"):
                    log(f"Update {file_to}")
                    if options.get("--dry-run"):
                        file_to_move[file_to] = False
                    else:
                        os.remove(file_to)
                elif options.get("--update"):
                    if is_run_script_file(f) and not options.get("--run-ops"):
                        if info_keep_run_ops:
                            info(f"Keep existing {file_to} (use -r to override)")
                        file_to_move[file_to] = False
                        continue
                    if is_test_file(f) and not options.get("--test-units"):
                        file_to_move[file_to] = False
                        continue
                    if is_user_file(f):
                        file_to_move[file_to] = False
                        # Have to update user analysis main file when switching frameworks
                        if is_framework_switch and is_analysis_user_file(mda_name, f):
                            file_to_move[file_to] = True
                        else:
                            continue
                    log(f"Update {file_to}")
                    if not options.get("--dry-run"):
                        os.remove(file_to)
                else:
                    warn(
                        f"File {file_to} in the way: remove it or use --force to override"
                    )
                    file_to_move[file_to] = False
            else:
                if options.get("--force"):
                    log(f"Pull {file_to}")
                    if options.get("--dry-run"):
                        file_to_move[file_to] = False
                elif options.get("--update") and (
                    (is_run_script_file(f) and not options.get("--run-ops"))
                    or (is_test_file(f) and not options.get("--test-units"))
                    or is_user_file(f)
                ):
                    file_to_move[file_to] = False
                else:
                    log(f"Pull {file_to}")

        if not options.get("--dry-run"):
            move_files(file_to_move, tempdir)
            state = {
                "whatsopt_url": self._url,
                "analysis_id": mda_id,
                "framework": framework,
                "pull_mode": MODE_PACKAGE if options.get("--package") else MODE_PLAIN,
            }
            save_state(state)
            info(msg)

    def pull_mda_json(self, mda_id):
        url = self.endpoint(f"/api/v1/analyses/{mda_id}.wopjson")
        resp = self.session.get(url, headers=self.headers, stream=True)
        WhatsOpt.check_http_error(resp)
        print(json.dumps(resp.json()))

    def pull_project_json(self, project_id):
        url = self.endpoint(f"/api/v1/design_projects/{project_id}.wopjson")
        resp = self.session.get(url, headers=self.headers, stream=True)
        WhatsOpt.check_http_error(resp)
        print(json.dumps(resp.json()))

    def update_mda(self, analysis_id=None, options={}, info_keep_run_ops=True):
        mda_id = analysis_id or get_analysis_id()
        if mda_id and not analysis_id:
            url = get_whatsopt_url()
            if url != self.url:
                warn(
                    f"You want to update code pulled from {url} while you are logged in {self.url}."
                )
                info(
                    f"  use 'wop login {url}' and retry to update using the original remote."
                )
                info(
                    f"  use 'wop update -a <id>' to update using analysis #<id> from {self.url}."
                )
                sys.exit(-1)
        if mda_id is None:
            error(
                f"Unknown analysis with id=#{mda_id} (maybe use wop pull <analysis-id>)"
            )
            sys.exit(-1)
        # keep options unchanged, work on a copy
        opts = copy.deepcopy(options)
        # sanity checks
        if not (is_based_on(FRAMEWORK_OPENMDAO) or is_based_on(FRAMEWORK_GEMSEO)):
            error("No framework detected. Check your *_base.py files.")
            sys.exit(-1)
        if opts["--openmdao"] and opts["--gemseo"]:
            error("Please choose either --openmdao or --gemseo.")
            sys.exit(-1)
        opts.update(
            {
                "--base": True,
                "--update": True,
                "--gemseo": opts["--gemseo"]
                or (not opts["--openmdao"] and is_based_on(FRAMEWORK_GEMSEO)),
                "--openmdao": opts["--openmdao"]
                or (not opts["--gemseo"] and is_based_on(FRAMEWORK_OPENMDAO)),
                "--package": is_package_mode(),
            }
        )
        self.pull_mda(
            mda_id, opts, "Analysis #{} updated".format(mda_id), info_keep_run_ops
        )

    def show_mda(self, analysis_id, pbfile, name, outfile, batch, depth):
        options = {
            "--xdsm": True,
            "--name": name,
            "--dry-run": False,
            "--depth": depth,
        }
        xdsm = None
        if pbfile:
            start = time.time()
            try:
                info("XDSM info retrieval...")
                self.push_mda_cmd(pbfile, options)
            except AnalysisPushedException as pushed:
                xdsm = pushed.xdsm
            end = time.time()
            log("Retrieved in {:.2f}s".format(end - start))
            source = os.path.basename(pbfile)
        else:
            mda_id = analysis_id or get_analysis_id()
            if mda_id is None:
                error(
                    "Unknown analysis with id={} (maybe use wop pull <analysis-id>)".format(
                        mda_id
                    )
                )
                sys.exit(-1)
            url = self.endpoint("/api/v1/analyses/{}.xdsm".format(mda_id))
            resp = self.session.get(url, headers=self.headers)
            WhatsOpt.check_http_error(resp)
            xdsm = resp.json()
            source = f"{mda_id}@{self._url}"

        info("XDSM building...")
        generate_xdsm_html(source, xdsm, outfile)
        if pbfile:
            log("XDSM of analysis from {} generated in {}".format(pbfile, outfile))
        else:
            log("XDSM of analysis {} generated in {}".format(mda_id, outfile))
        if not batch:
            webview(outfile)

    def upload(
        self,
        filename,
        driver_kind=None,
        analysis_id=None,
        operation_id=None,
        dry_run=False,
        outvar_count=1,
        only_success=False,
        parallel=False,
    ):
        from socket import gethostname

        mda_id = get_analysis_id() if not analysis_id else analysis_id
        # Test sqlite files generated with MPI
        _, extension = os.path.splitext(filename)
        parallel_sqlite = re.match(r"\.sqlite_\d+$", extension)

        name = cases = statuses = None
        if not os.path.exists(filename):
            error(f"File not found ({filename})")
            sys.exit(-1)
        if os.path.basename(filename) == "mda_init.py":
            self.upload_vars_init_cmd(
                filename, {"--dry-run": dry_run, "--analysis-id": mda_id}
            )
        elif filename.endswith(".csv"):
            name, cases, statuses = load_from_csv(filename)
        elif filename.endswith(".sqlite") or (parallel_sqlite and parallel):
            name, cases, statuses = load_from_sqlite(filename, parallel)
        elif filename.endswith(".hdf5"):
            name, cases, statuses = load_from_hdf5(filename)
        else:
            error(
                f"Can not upload file {filename}: extension not recognized"
                " (should be either .csv, .sqlite or .hdf5)"
            )
            sys.exit(-1)

        if only_success:
            for c in cases:
                c["values"] = [
                    val for i, val in enumerate(c["values"]) if statuses[i] > 0
                ]
            statuses = [1 for s in statuses if s > 0]

        for c in cases:
            c["values"] = np.nan_to_num(np.array(c["values"])).tolist()

        if dry_run:
            print_cases(cases, statuses)
            sys.exit()

        resp = None
        if operation_id:
            url = self.endpoint(("/api/v1/operations/%s") % operation_id)
            operation_params = {"cases": cases}
            resp = self.session.patch(
                url, headers=self.headers, json={"operation": operation_params}
            )
        else:
            if mda_id:
                url = self.endpoint(("/api/v1/analyses/%s/operations") % mda_id)
            else:
                url = self.endpoint("/api/v1/operations")
            if driver_kind:
                driver = "user_{}_algo".format(driver_kind)
            else:
                # suppose name well-formed <lib>-<doe|optimizer|screening>-<algoname>
                # otherwise it will default to doe
                m = re.match(r"(\w+)_(doe|optimizer|screening)_(\w+)", name.lower())
                if m:
                    driver = name.lower()
                else:
                    driver = "user_data_uploading"
            operation_params = {
                "name": name,
                "driver": driver,
                "host": gethostname(),
                "cases": cases,
                "success": statuses,
            }
            params = {"operation": operation_params}
            if outvar_count > 0 and outvar_count < len(cases):
                params["outvar_count_hint"] = outvar_count
            resp = self.session.post(url, headers=self.headers, json=params)
        WhatsOpt.check_http_error(resp)
        log("Results data from {} uploaded with driver {}".format(filename, driver))
        if mda_id:
            log(f"attached to analysis #{mda_id}")

    def upload_vars_init_cmd(self, py_filename, options):
        def upload_vars_init(prob):
            self.upload_vars_init(prob, options)
            sys.exit()

        d = os.path.dirname(py_filename)
        run_mda_filename = os.path.join(d, "run_mda.py")
        if not os.path.exists(run_mda_filename):
            error(f"Can not get analysis init: script {run_mda_filename} not found.")
        hooks.use_hooks = True
        hooks._register_hook("final_setup", "Problem", post=upload_vars_init)
        _load_and_exec(run_mda_filename, [])

    def upload_vars_init(self, problem, options):
        mda_id = get_analysis_id() if get_analysis_id() else options["--analysis-id"]
        if mda_id is None:
            error("Unknown analysis with id={}".format(mda_id))
            sys.exit(-1)
        parameters = []
        headers = ["variable", "init value"]
        data = []
        for s in problem.model._subsystems_myproc:
            if isinstance(s, IndepVarComp):
                for absname in s._var_abs2meta["output"]:
                    name = find_indep_var_name(problem, absname)
                    value = None
                    if Version(OPENMDAO_VERSION) < Version("3.39"):
                        value = s._outputs._views[absname][0]
                    else:
                        value = s._outputs.get_val(absname, flat=False)
                    if isinstance(value, np.ndarray):
                        value = str(value.tolist())
                    parameters.append({"varname": name, "value": value})
        data = [[p["varname"], p["value"]] for p in parameters]
        params = {"parameterization": {"parameters": parameters}}
        log(tabulate(data, headers))
        if not options["--dry-run"]:
            url = self.endpoint(f"/api/v1/analyses/{mda_id}/parameterization")
            resp = self.session.put(url, headers=self.headers, json=params)
            WhatsOpt.check_http_error(resp)
            log("Variables init values uploaded")

    def check_versions(self):
        url = self.endpoint("/api/v1/versioning")
        resp = self.session.get(url, headers=self.headers)
        WhatsOpt.check_http_error(resp)
        version = resp.json()
        log("WhatsOpt {} requires wop {}".format(version["whatsopt"], version["wop"]))
        log(f"You are using wop {__version__}")

    @staticmethod
    def serve(port):
        try:
            import thrift  # noqa: F401
        except ImportError:
            error(
                "Apache Thrift is not installed. You can install it with : 'pip install thrift'"
            )
            exit(-1)
        try:
            import os
            import sys

            # insert current dir first as another run_server exists under whatsopt/services
            sys.path.insert(0, os.getcwd())
            from run_server import run_server
        except ImportError as err:
            print(str(err))
            error("Server not found!")
            try:
                mda_id = get_analysis_id()
                if mda_id:
                    log(
                        f"  (use 'wop update -s' to generate server for current analysis #{mda_id})"
                    )
                else:
                    warn("No local analysis found")
                    log(
                        "  (use 'wop pull -s <id>' to generate server for the analysis #id)"
                    )
            except ValueError as err:
                warn(str(err))
            sys.exit(-1)
        run_server(port)

    @staticmethod
    def convert(filename):
        if not os.path.exists(filename):
            error(f"File {filename} not found.")
        pathname, extension = os.path.splitext(filename)
        if not (extension == ".sqlite" or re.match(r"\.sqlite_\d+$", extension)):
            warn(
                f"File {filename} should have '.sqlite[_n]' extension, got '{extension}'"
            )
        basename = os.path.basename(pathname)
        convert_sqlite_to_csv(filename, basename)

    def publish(self, force=False, analysis_id=None):
        mda_id = analysis_id if analysis_id else get_analysis_id()
        url = self.endpoint(f"/api/v1/analyses/{mda_id}/package")
        if not force:
            resp = self.session.get(url, headers=self.headers)
            if resp.ok:
                existing_meta = resp.json()
            elif resp.status_code == HTTPStatus.NOT_FOUND:
                existing_meta = None
            else:
                WhatsOpt.check_http_error(resp)

            if existing_meta:
                warn(
                    f"Existing package {existing_meta['name']} {existing_meta['version']} will be lost."
                )
                answer = input("Do you want to continue? (yes/no): ")
                if answer.lower() == "yes":
                    pass
                else:
                    info("Publishing aborted")
                    exit(-1)

        info("Package building...")
        filename = self.build()
        if not os.path.exists(filename):
            error(f"File {filename} not found")
            exit(-1)
        meta = get_pkg_metadata(filename)
        if (
            not force
            and existing_meta
            and existing_meta["name"] == meta.name
            and Version(existing_meta["version"]) >= Version(meta.version)
        ):
            error(f"You have to bump the version (> {existing_meta['version']})")
            error("Publishing aborted")
            exit(-1)

        files = {
            "package[description]": (None, meta.summary, "application/json"),
            "package[archive]": (filename, open(filename, "rb"), "application/gzip"),
        }
        info("Package publishing...")
        resp = self.session.post(url, headers=self.headers, files=files)
        if resp.ok:
            info(
                f"Package {meta.name} v{meta.version} is published on WopStore({self.endpoint('/packages')})"
            )
        elif resp.status_code == HTTPStatus.UNPROCESSABLE_ENTITY:
            error(resp.json()["message"])
            error("Duplicate detected! The package already exists on WopStore!")
            exit(-1)
        else:
            WhatsOpt.check_http_error(resp)

    def build(self):
        if not is_package_mode():
            error("Package mode is required!")
            exit(-1)
        self._update_mda_base()
        filename = build_package()
        return filename

    def fetch(self, source_id=None, options={}):
        if not is_package_mode():
            error("Package mode is required!")
            exit(-1)
        mda_id = get_analysis_id()
        param = f"?src_id={source_id}"
        format_query = "mda_pkg_content"
        url = self.endpoint(
            f"/api/v1/analyses/{mda_id}/exports/new.{format_query}{param}"
        )
        resp = self.session.get(url, headers=self.headers, stream=True)
        if resp.ok:
            name = None
            with tempfile.NamedTemporaryFile(
                suffix=".zip", mode="wb", delete=False
            ) as fd:
                for chunk in resp.iter_content(chunk_size=128):
                    fd.write(chunk)
                name = fd.name
            zipf = zipfile.ZipFile(name, "r")
            tempdir = tempfile.mkdtemp(suffix="wop", dir=tempfile.tempdir)
            zipf.extractall(tempdir)
            filenames = zipf.namelist()
            if not filenames:
                warn(
                    f"Fetching but no package found for analysis #{source_id}, nothing to do"
                )
                return
            zipf.close()

            file_to_move = {}
            for f in filenames:
                file_to = f
                file_to_move[file_to] = True
                if os.path.exists(file_to):
                    if options.get("--force"):
                        log(f"Fetch {file_to}")
                        if options.get("--dry-run"):
                            file_to_move[file_to] = False
                        else:
                            os.remove(file_to)
                    else:
                        warn(
                            f"File {file_to} in the way: remove it or use --force to override"
                        )
                        file_to_move[file_to] = False
                else:
                    log(f"Fetch {file_to}")
            if not options.get("--dry-run"):
                move_files(file_to_move, tempdir)
                info(f"Analysis #{source_id} disciplines fetched")
        else:
            error(f"Error while fetching disciplines of analysis #{source_id}")
            error(resp.json().get("message"))

    def merge(self, source_id, options={}):
        if not is_package_mode():
            error("Package mode is required!")
            exit(-1)
        current_id = get_analysis_id()
        if int(source_id) == current_id:
            warn(f"Merging analysis #{current_id} in itself, nothing to do")
            return
        url = self.endpoint(f"/api/v1/analyses/{current_id}")
        params = {
            "analysis": {
                "import": {"analysis": source_id},
            },
            "requested_at": str(datetime.now()),
        }
        if options.get("--dry-run"):
            self.get_status()
            url = self.endpoint("/api/v1/analyses/{}".format(source_id))
            resp = self.session.get(url, headers=self.headers)
            log(f"Analysis #{source_id} #{resp.json()} is selected to be merged")
        else:
            resp = self.session.put(url, headers=self.headers, json=params)
        if resp.ok:
            if not options.get("--dry-run"):
                info(f"Analysis #{source_id} merged")
        elif resp.status_code == HTTPStatus.UNPROCESSABLE_ENTITY:
            error(f"Error while merging analysis #{source_id}.")
            error(
                "    Check analyses, maybe they are not compatible (same variable produced by different disciplines)"
            )
        elif resp.status_code == HTTPStatus.FORBIDDEN:
            error(f"Error while merging analysis #{source_id}.")
            error(
                "    You are not authorized to update the current analysis: either you do not own it or"
                " current analysis is already packaged or operated"
            )
        else:
            error(f"Error while merging analysis #{source_id}")
            WhatsOpt.check_http_error(resp)

    def pull_source_mda(self, source_id, options={}):
        self.merge(source_id, options)
        self.fetch(source_id, options)
        self._update_mda_base(options.get("--dry-run"))

    def _update_mda_base(self, dry_run=False, force=False):
        update_options = {
            "--dry-run": dry_run,
            "--force": force,
            "--server": False,
            "--egmdo": False,
            "--run-ops": False,
            "--test-units": False,
            "--gemseo": is_based_on(FRAMEWORK_GEMSEO),
            "--openmdao": is_based_on(FRAMEWORK_OPENMDAO),
        }
        mda_id = get_analysis_id()
        self.update_mda(mda_id, update_options, info_keep_run_ops=False)

    def _test_connection(self):
        if self.api_key:
            self.headers = {
                "Authorization": "Token token=" + self.api_key,
                "User-Agent": "wop/{}".format(__version__),
            }
            url = self.endpoint("/api/v1/versioning")
            debug(f"Test connect: {url}, {self.api_key}")
            try:
                resp = self.session.get(url, headers=self.headers)
                # special case: bad wop version < minimal required version
                if resp.status_code == requests.codes.forbidden:
                    error(resp.json()["message"])
                    sys.exit(-1)
                return resp.ok
            except requests.exceptions.ConnectionError:
                return False
        else:
            return False

    @staticmethod
    def check_http_error(resp):
        msg = None
        try:
            msg = resp.json().get("message")
        except requests.exceptions.JSONDecodeError:
            pass  # Does not contain json => ignore
        if msg:
            warn(msg)
        try:
            resp.raise_for_status()
        except requests.exceptions.HTTPError as http_err:
            error(f"HTTP Error : {http_err}")
            exit(-1)



import os
import re
import sys
import tomli_w
from whatsopt.logging import error
from shutil import move


WOP_CONF_FILENAME = ".wop"

WOP_FORMAT_VERSION_KEY = "wop_format_version"
WHATSOPT_URL_KEY = "whatsopt_url"
ANALYSIS_ID_KEY = "analysis_id"
FRAMEWORK_KEY = "framework"
PULL_MODE_KEY = "pull_mode"

FRAMEWORK_OPENMDAO = "openmdao"
FRAMEWORK_GEMSEO = "gemseo"

MODE_PLAIN = "plain"
MODE_PACKAGE = "package"


def save_state(
    state,
    filename=WOP_CONF_FILENAME,
):
    comment = """# This file contains recorded state from wop pull/update commands
# DO NOT EDIT unless you know what you are doing
# Version history:
# * version 2: use toml format, add wop_format_version
# * version 1: initial format "key: val"
# * version 0: no wop file
#
"""
    _state = {}
    _state[WOP_FORMAT_VERSION_KEY] = 2
    _state[WHATSOPT_URL_KEY] = state[WHATSOPT_URL_KEY]
    _state[ANALYSIS_ID_KEY] = int(state[ANALYSIS_ID_KEY])
    _state[FRAMEWORK_KEY] = state[FRAMEWORK_KEY]
    _state[PULL_MODE_KEY] = state[PULL_MODE_KEY]
    content = tomli_w.dumps(_state)

    with open(filename, "w") as f:
        f.write(comment)
        f.write(content)


def load_state(filename=WOP_CONF_FILENAME):
    # Should be able to load version 0, version 1 and version 2 format
    state = {"wop_format_version": 0}
    if not os.path.exists(filename):
        return state
    with open(filename, "r") as f:
        for line in f.readlines():
            line = line.strip()
            if line == "" or line.startswith("#"):
                continue
            m = re.search(r"(\S+)\s*[:=]\s*(\S+)", line)
            if m:
                val = m.group(2)
                if val.startswith('"') or val.startswith("'"):
                    val = val[1:-1]
                if re.match(r"\d+", val):
                    val = int(val)

                state[m.group(1)] = val
            else:
                error(f"Syntax error in {filename} file: line '{line}' invalid")
                sys.exit(-1)
    if not state.get("wop_format_version"):
        state["wop_format_version"] = 1

    return state


def is_user_file(f):
    return (
        not re.match(r".*_base\.py$", f)
        and not is_run_script_file(f)
        and not is_test_file(f)
        and not re.match(r"^.*server/", f)
        and not re.match(r"^.*egmdo/", f)
    )


def is_run_script_file(f):
    return f == "mda_init.py" or re.match(r"^run_.*\.py$", f)


def is_test_file(f):
    return re.match(r"^.*tests/test_.*\.py$", f)


def is_analysis_user_file(name, f):
    return bool(re.match(rf"^{name}\.py$", f))


def find_analysis_base_files(directory="."):
    files = []
    for f in os.listdir(directory):
        if f.endswith("_base.py"):
            files.append(f)
    return files


def _extract_key(file, key):
    ident = None
    with open(file, "r") as f:
        pattern = re.compile(rf"^# {key}: (.*)")
        for line in f:
            match = pattern.match(line)
            if match:
                ident = match.group(1)
                break
    return ident


def extract_remote_name(url_or_name):
    m = re.match(r"https?:\/\/(\w+)", url_or_name)
    if m:  # url
        return m.group(1)
    elif re.match(r"^\w+$", url_or_name):  # name
        return url_or_name
    else:
        print(f"Warning: Can not find remote name out of {url_or_name}")
        return None


def _get_key(key, directory="."):
    files = find_analysis_base_files(directory)
    val = None
    for f in files:
        newval = _extract_key(os.path.join(directory, f), key)
        if val and val != newval:
            raise ValueError(
                "Several {} key detected. "
                "Find #{} then #{}.\n"
                "Check header comments in {} files.".format(
                    key, val, newval, str(files)
                )
            )
        val = newval
    return val


def get_analysis_id(directory="."):
    state = load_state()
    if state["wop_format_version"] > 0:
        return state[ANALYSIS_ID_KEY]
    else:
        return _get_key(ANALYSIS_ID_KEY, directory)


def get_whatsopt_url(directory="."):
    state = load_state()
    if state["wop_format_version"] > 0:
        return state[WHATSOPT_URL_KEY]
    else:
        try:
            return _get_key(WHATSOPT_URL_KEY, directory)
        except ValueError:
            return False


def is_based_on(module, directory="."):
    state = load_state()
    if state["wop_format_version"] > 0:
        return state[FRAMEWORK_KEY] == module
    else:
        files = find_analysis_base_files(directory)
        return len(files) > 0 and all(
            _detect_from_import(os.path.join(directory, f), module) for f in files
        )


def is_framework_switch(framework, directory="."):
    return (
        framework == FRAMEWORK_GEMSEO and is_based_on(FRAMEWORK_OPENMDAO, directory)
    ) or (framework == FRAMEWORK_OPENMDAO and is_based_on(FRAMEWORK_GEMSEO, directory))


def _detect_from_import(file, module):
    detected = False
    with open(file, "r") as f:
        for line in f:
            # TODO: Maybe would need more robust detection... We'll see!
            # first from/import detected gives the framework.
            match = re.match(rf"^(from|import) {module}\..*", line)
            if match:
                detected = True
                break
    return detected


def is_package_mode():
    state = load_state()
    return state and state[PULL_MODE_KEY] == MODE_PACKAGE


def move_files(file_to_move, tempdir):
    for f in file_to_move.keys():
        file_from = os.path.join(tempdir, f)
        file_to = f
        dir_to = os.path.dirname(f)
        if dir_to == "":
            dir_to = "."
        elif not os.path.exists(dir_to):
            os.makedirs(dir_to)
        if file_to_move[file_to]:
            move(file_from, dir_to)
